"""Class for generating tables from a list of file uploads"""

import pathlib
import re
import sys

import msgspec
import pandas
import platformdirs

from cumulus_library import BaseTableBuilder, base_utils, errors, study_manifest
from cumulus_library.template_sql import base_templates

"""
A file upload config looks like this

type="file_upload"
[tables]
[table.name] # the table in db will be called 'name', snaked cased for sql compatibility
files = ["relative/path/from/config/file.csv"]
# The rest of these are optional
delimiter = "|" # comma by default, only applies to csv/tsv/bsv
col_types =["STRING","DATE","DOUBLE"] # default is all strings. Only relevant for Athena.
# See https://cwiki.apache.org/confluence/display/Hive/LanguageManual+Types for valid types
[table.other_name]
file = "other/path/file.csv"
"""


class FileUploadTask(msgspec.Struct, forbid_unknown_fields=True):
    file: str | None = None  # deprecated
    files: list[str] | None = None
    create_mode: str = "single"
    delimiter: str | None = None
    col_types: list[str] | None = None
    always_upload: bool = False


class FileUploadWorkflow(msgspec.Struct, forbid_unknown_fields=True):
    config_type: str
    tables: dict[str, FileUploadTask]


class FileUploadBuilder(BaseTableBuilder):
    display_text = "Uploading files..."

    @staticmethod
    def _snake_case(x: str) -> str:
        """Converts either CAPS_CASE or CamelCase to snake_case"""
        return re.sub(r"([a-z])([A-Z])", r"\1_\2", x).lower()

    def __init__(self, *args, toml_config_path: pathlib.Path, **kwargs):
        super().__init__(*args, **kwargs)
        self._toml_config_path = toml_config_path
        self._toml_config_dir = toml_config_path.parent
        try:
            with open(toml_config_path, "rb") as file:
                file_bytes = file.read()
                self._workflow_config = msgspec.to_builtins(
                    msgspec.toml.decode(file_bytes, type=FileUploadWorkflow)
                )
        except msgspec.ValidationError as e:  # pragma: no cover
            sys.exit(
                f"The file upload workflow at {toml_config_path!s} contains an "
                "unexpected param: \n"
                f"{e}"
            )
        for table in self._workflow_config["tables"]:
            # migrate deprecated `file` key if found
            if file := self._workflow_config["tables"][table]["file"]:
                self._workflow_config["tables"][table]["files"] = [file]
            if mode := self._workflow_config["tables"][table]["create_mode"] not in (
                "single",
                "multiple",
            ):
                raise errors.FileUploadError(  # pragma: no cover
                    f"Create mode '{mode}' in file upload workflow "
                    f"{self._toml_config_path} is invalid."
                )

    def prepare_queries(
        self,
        config: base_utils.StudyConfig,
        manifest: study_manifest.StudyManifest,
        *args,
        **kwargs,
    ):
        cache_dir = (
            pathlib.Path(platformdirs.user_cache_dir("cumulus-library", "smart-on-fhir"))
            / f"file_uploads/{manifest.get_study_prefix()}"
        )
        cache_dir.mkdir(parents=True, exist_ok=True)

        with base_utils.get_progress_bar() as progress:
            task = progress.add_task(
                "Uploading static files...",
                total=len(self._workflow_config["tables"]),
            )
            for task_name in self._workflow_config["tables"]:
                table = self._workflow_config["tables"][task_name]

                # we want to generate one query per task - if we hit a case where
                # multiple files are backing a similar table, we'll skip appending
                # copies of the same query later on.
                generated_query = False
                if table["create_mode"] == "single":
                    # For single table mode with multiple files, we'll track the dataframe
                    # column types after the first file is ingested, and throw an error if
                    # we have a mismatch, rather than waiting for the error to happen at the
                    # SQL level. If there's one file, this will get set but then never used
                    # for validation
                    types = None

                # To handle cases where we're using a folder of data to back either a single
                # table with multiple files, or creation of multiple tables, we'll scan for
                # directories and replace them with their contents
                new_files = []
                for file_or_dir in table["files"]:
                    if (self._toml_config_dir / file_or_dir).is_dir():
                        dir_contents = [
                            f"{file_or_dir}/{x.name}"
                            for x in (self._toml_config_dir / file_or_dir).glob("**/*")
                        ]
                        new_files = new_files + dir_contents
                    else:
                        new_files.append(file_or_dir)
                table["files"] = new_files
                for file in table["files"]:
                    table_filename = pathlib.Path(file).name
                    if file.endswith(".md"):
                        continue
                    elif file.endswith(".xlsx"):
                        parquet_path = cache_dir / table_filename.replace(".xlsx", ".parquet")
                        df = pandas.read_excel(self._toml_config_dir / file)

                    elif file.endswith(".csv") or file.endswith(".tsv") or file.endswith(".bsv"):
                        parquet_path = cache_dir / f"{table_filename[:-4]}.parquet"
                        if table["delimiter"] is None:
                            match file.split(".")[-1]:
                                case "bsv":
                                    table["delimiter"] = "|"
                                case "csv":
                                    table["delimiter"] = ","
                                case "tsv":
                                    table["delimiter"] = "\t"
                        df = pandas.read_csv(
                            self._toml_config_dir / file, delimiter=table["delimiter"]
                        )

                    elif file.endswith(".parquet"):
                        parquet_path = cache_dir / table_filename
                        df = pandas.read_parquet(self._toml_config_dir / file)

                    else:
                        raise errors.FileUploadError(
                            f"{table['file']} is not a supported upload file type.\n"
                            "Supported file types: csv, bsv, tsv, xlsx, parquet"
                        )
                    if table["create_mode"] == "single":
                        if types is None:
                            types = df.dtypes
                        elif list(types.items()) != list(df.dtypes.items()):
                            raise errors.FileUploadError(
                                f"{table_filename} does not match the schema of the other files"
                            )
                        # for single table mode, we can use the name of the task from the workflow
                        table_name = task_name
                    else:
                        # for multiple tables mode, we'll want to infer it from the filename
                        table_name = pathlib.Path(file).stem
                    path_parts = list(parquet_path.parts)
                    path_parts.insert(-1, table_name)
                    parquet_path = pathlib.Path().joinpath(*path_parts)
                    local_location = parquet_path.parent

                    df = df.rename(self._snake_case, axis="columns")
                    if table["col_types"] is None:
                        table["col_types"] = ["STRING" for x in df.columns]
                    if len(table["col_types"]) != len(df.columns):
                        raise errors.FileUploadError(
                            f"{task_name} has {len(df.columns)} columns, but the provided "
                            f"col_types has {len(table['col_types'])} entries."
                        )
                    type_dict = {}
                    for pos in range(0, len(table["col_types"])):
                        type_dict[df.columns[pos]] = table["col_types"][pos].lower()
                    df = df.astype(type_dict)
                    parquet_path.parent.mkdir(parents=True, exist_ok=True)
                    df.to_parquet(parquet_path)
                    remote_path = config.db.upload_file(
                        file=parquet_path,
                        study=manifest.get_study_prefix(),
                        topic=table_name,
                        force_upload=table["always_upload"] or config.force_upload,
                    )
                    if not generated_query or table["create_mode"] == "multiple":
                        self.queries.append(
                            base_templates.get_ctas_from_parquet_query(
                                schema_name=config.schema,
                                table_name=f"{manifest.get_study_prefix()}__{table_name}",
                                local_location=local_location,
                                remote_location=remote_path,
                                table_cols=df.columns,
                                remote_table_cols_types=table["col_types"],
                            )
                        )
                        generated_query = True
                progress.advance(task)



import sys

import rich
from rich import progress

from cumulus_library import base_utils, databases, enums, errors, study_manifest
from cumulus_library.template_sql import base_templates


def _execute_drop_queries(
    db: databases.DatabaseBackend,
    verbose: bool,
    view_table_list: list,
    progress: progress.Progress,
    task: progress.TaskID,
) -> None:
    """Handler for executing drop view/table queries and displaying console output.

    :param cursor: A DatabaseCursor object
    :param verbose: toggle from progress bar to query output
    :param view_table_list: a list of views and tables beginning with
      the study prefix
    :param progress: a rich progress bar renderer
    :param task: a TaskID for a given progress bar
    """
    queries = []
    for view_table in view_table_list:
        queries.append(
            base_templates.get_drop_view_table(name=view_table[0], view_or_table=view_table[1])
        )
    db.parallel_write(queries, verbose, progress, task)


def _get_unprotected_stats_view_table(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest,
    drop_prefix: str,
    study_prefix: str,
    clean_by_cli_prefix: bool,
    query: str | None = None,
    artifact_type: str | None = None,
    artifact_list: str | None = None,
):
    """Gets all items from the database by type, less any protected items

    :param config: A StudyConfig object
    :param manifest: A StudyManifest object
    :param drop_prefix: The prefix requested to drop
    :param study_prefix: The expected study prefix
    :param clean_by_cli_prefix: if True, `--prefix` was passed in as an arg
    :keyword query: A query to get the raw list of items from the db
    :keyword artifact_type: either 'table' or 'view'
    :keyword artifact_list: a list of tuples, contained (query, artifact_type) pairs

    :returns: a list of study tables to drop
    """
    cursor = config.db.cursor()
    unprotected_list = []
    if query and artifact_type:
        db_contents = cursor.execute(query).fetchall()
    elif artifact_list:
        db_contents = artifact_list
    else:  # pragma: no cover
        raise errors.CumulusLibraryError(
            "cleaner._get_unprotected_stats_view_table() requires one of the following is true:\n"
            "  - 'query' and 'artifact_type' are provided\n"
            "  - 'artifact_list' is provided"
        )
    if manifest.has_stats() and not config.stats_clean and not clean_by_cli_prefix:
        protected_list = cursor.execute(
            f"""SELECT table_name
            FROM {drop_prefix}{enums.ProtectedTables.STATISTICS.value}
            WHERE study_name = '{study_prefix}'"""  # noqa: S608
        ).fetchall()
        for protected_tuple in protected_list:
            db_contents = list(filter(lambda x: x[0] != protected_tuple[0], db_contents))
    for db_row_tuple in db_contents:
        # this check handles athena reporting views as also being tables,
        # so we don't waste time dropping things that don't exist
        if artifact_list:
            if not any(db_row_tuple[0] in iter_q_and_t for iter_q_and_t in unprotected_list):
                unprotected_list.append([db_row_tuple[0], db_row_tuple[1]])
        elif artifact_type == "TABLE":
            if not any(db_row_tuple[0] in iter_q_and_t for iter_q_and_t in unprotected_list):
                unprotected_list.append([db_row_tuple[0], artifact_type])
        else:
            unprotected_list.append([db_row_tuple[0], artifact_type])
    return unprotected_list


def clean_study(
    config: base_utils.StudyConfig,
    manifest: study_manifest.StudyManifest | None,
    prefix: str | None = None,
) -> list:
    """Removes tables beginning with the study prefix from the database schema

    :param config: a StudyConfig object
    :param manifest: a StudyManifest object
    :keyword prefix: override manifest-based prefix discovery with the provided prefix
    :returns: list of dropped tables (for unit testing only)

    """
    if not prefix and not manifest:
        raise errors.CumulusLibraryError(
            "Either a manifest parser or a filter prefix must be provided"
        )
    if prefix:
        drop_prefix = prefix
        study_prefix = prefix
    else:
        drop_prefix = manifest.get_schema_aware_prefix_with_seperator()
        study_prefix = manifest.get_schema_aware_prefix()

    if config.stats_clean:
        # Note: we don't expect dedicated schema studies to have stats,
        # so this prompt does not need to handle the case where the study
        # prefix is an empty string
        confirm = input(
            "This will remove all historical stats tables in the "
            f"{study_prefix} study - are you sure? (y/N)"
        )
        if confirm is None or confirm.lower() not in ("y", "yes"):
            sys.exit("Table cleaning aborted")

    cursor = config.db.cursor()

    view_table_list = []
    if config.stage != "all" and not prefix:
        # We're going to inspect the list of tables to see if we've got a build source
        # table for the study being cleaned. If we don't, than the study was last built
        # with a pre-V6 version of the library, and we'll fall back to prefix cleaning
        # mode so a user doesn't have to manage migration themselves.
        query = base_templates.get_select_from_single_query(
            schema="information_schema",
            table_name="tables",
            columns=["table_name"],
            where_clauses=[
                f"table_name = '{drop_prefix}{enums.ProtectedTables.BUILD_SOURCE.value}'\n"
                f"AND table_schema = '{config.schema}'"
            ],
        )
        res = cursor.execute(query).fetchall()
        if len(res) == 0:
            prefix = manifest.get_study_prefix()
            drop_prefix = prefix
            study_prefix = prefix

        else:
            query = base_templates.get_select_from_single_query(
                schema=base_utils.get_schema(config=config, manifest=manifest),
                table_name=f"{drop_prefix}{enums.ProtectedTables.BUILD_SOURCE.value}",
                columns=["name", "type"],
                where_clauses=[f"stage = '{config.stage}'"],
                distinct=True,
            )
            names_and_types = cursor.execute(query).fetchall()
            if names_and_types != []:
                view_table_list += _get_unprotected_stats_view_table(
                    config,
                    manifest,
                    drop_prefix,
                    study_prefix,
                    clean_by_cli_prefix=isinstance(prefix, str),
                    artifact_list=names_and_types,
                )
    if config.stage == "all" or prefix:
        view_sql = base_templates.get_show_views(config.schema, drop_prefix)
        table_sql = base_templates.get_show_tables(config.schema, drop_prefix)
        for query, artifact_type in [[view_sql, "VIEW"], [table_sql, "TABLE"]]:
            view_table_list += _get_unprotected_stats_view_table(
                config,
                manifest,
                drop_prefix,
                study_prefix,
                clean_by_cli_prefix=isinstance(prefix, str),
                query=query,
                artifact_type=artifact_type,
            )

    if not view_table_list:
        return view_table_list
    # We'll do a pass to see if any of these tables are protected, and remove them from the list.
    for view_table in view_table_list.copy():
        if any(
            (
                f"__{word.value}_" in view_table[0]
                or view_table[0].endswith(f"__{word.value}")
                or view_table[0].startswith(f"{word.value}_")
            )
            for word in enums.ProtectedTableKeywords
        ):
            view_table_list.remove(view_table)
    if prefix:
        rich.print("The following views/tables were selected by prefix:")
        for view_table in view_table_list:
            rich.print(f"  {view_table[0]}")
        confirm = input("Remove these tables? (y/N)")
        if confirm is None or confirm.lower() not in ("y", "yes"):
            sys.exit("Table cleaning aborted")
    if dedicated := manifest.get_dedicated_schema():
        view_table_list = [
            (
                # Athena uses different quoting strategies for drop view statements
                # versus drop table statements. -_-
                # TODO: Consider moving this logic to a database object?
                f"`{dedicated}`.`{x[0]}`"
                if (x[1] == "TABLE" and config.db.db_type == "athena")
                else f'"{dedicated}"."{x[0]}"',
                x[1],
            )
            for x in view_table_list
        ]
    # We want to only show a progress bar if we are :not: printing SQL lines
    with base_utils.get_progress_bar(disable=config.verbose) as progress:
        task = progress.add_task(
            f"Removing {study_prefix} study artifacts...",
            total=len(view_table_list),
            visible=not config.verbose,
        )
        _execute_drop_queries(
            config.db,
            config.verbose,
            view_table_list,
            progress,
            task,
        )
    # if we're doing a stats clean, we'll also remove the table containing the
    # list of protected tables
    if config.stats_clean:
        drop_query = base_templates.get_drop_view_table(
            f"{drop_prefix}{enums.ProtectedTables.STATISTICS.value}", "TABLE"
        )
        cursor.execute(drop_query)

    if prefix is None:
        cleanup_query = base_templates.get_delete_from_table_query(
            schema=base_utils.get_schema(config=config, manifest=manifest),
            table_name=f"{drop_prefix}{enums.ProtectedTables.BUILD_SOURCE.value}",
            where_clauses=[f"stage = '{config.stage}'"],
        )
        cursor.execute(cleanup_query)

    return view_table_list
